import functools
from typing import Type
import klayout.dbcore
from klayout.dbcore import *

from klayout.db.pcell_declaration_helper import PCellDeclarationHelper

__all__ = klayout.dbcore.__all__ + ["PCellDeclarationHelper"]  # type: ignore

# Implementing deepcopy of common objects
# Point-like classes
PointLike = (Point, DPoint, DVector, Vector)


def pyaPoint__deepcopy__(self, memo):
    return self.dup()


def convert_type_error_to_not_implemented(cls, method):
    """If cls.method exists raises a TypeError, patch it so
    it returns a NotImplemented error instead.


    """
    if not hasattr(cls, method):
        return

    old_func = getattr(cls, method)

    @functools.wraps(old_func)
    def new_func(*args, **kwargs):
        try:
            return old_func(*args, **kwargs)
        except TypeError:
            return NotImplemented
    try:
        setattr(cls, method, new_func)
    except TypeError:
        # Some classes are immutable and cannot be changed.
        # At the time of writing, this happens to (_StaticAttribute, _AmbiguousMethodDispatcher, _Iterator, _Signal).__or__
        return

for PClass in PointLike:
    PClass.__deepcopy__ = pyaPoint__deepcopy__  # type: ignore

for cls in klayout.dbcore.__dict__.values():
    if not isinstance(cls, type):  # skip if not a class
        continue
    for method in (
        "__add__",
        "__sub__",
        "__mul__",
        "__matmul__",
        "__truediv__",
        "__floordiv__",
        "__mod__",
        "__divmod__",
        "__pow__",
        "__lshift__",
        "__rshift__",
        "__and__",
        "__xor__",
        "__or__",
    ):
        # list of methods extracted from https://docs.python.org/3.7/reference/datamodel.html#emulating-numeric-types
        convert_type_error_to_not_implemented(cls, method)


# If class has from_s, to_s, and assign, use them to
# enable serialization.
for name, cls in klayout.dbcore.__dict__.items():
    if not isinstance(cls, type):
        continue
    if hasattr(cls, 'from_s') and hasattr(cls, 'to_s') and hasattr(cls, 'assign'):
        cls.__getstate__ = cls.to_s  # type: ignore
        def _setstate(self, str):
            cls = self.__class__
            self.assign(cls.from_s(str))
        cls.__setstate__ = _setstate  # type: ignore



from typing import Any, ClassVar, Dict, Sequence, List, Iterator, Optional
from typing import overload
import klayout.db as db
class RdbReference:
    r"""
    @brief A cell reference inside the report database
    This class describes a cell reference. Such reference object can be attached to cells to describe instantiations of them in parent cells. Not necessarily all instantiations of a cell in the layout database are represented by references and in some cases there might even be no references at all. The references are merely a hint how a marker must be displayed in the context of any other, potentially parent, cell in the layout database.
    """
    parent_cell_id: int
    r"""
    @brief Gets parent cell ID for this reference
    @return The parent cell ID
    @brief Sets the parent cell ID for this reference
    """
    trans: db.DCplxTrans
    r"""
    @brief Gets the transformation for this reference
    The transformation describes the transformation of the child cell into the parent cell. In that sense that is the usual transformation of a cell reference.
    @return The transformation
    @brief Sets the transformation for this reference
    """
    def __copy__(self) -> RdbReference:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self, trans: db.DCplxTrans, parent_cell_id: int) -> None:
        r"""
        @brief Creates a reference with a given transformation and parent cell ID
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: RdbReference) -> None:
        r"""
        @brief Assigns another object to self
        """
    def database(self) -> ReportDatabase:
        r"""
        @brief Gets the database object that category is associated with

        This method has been introduced in version 0.23.
        """
    def dup(self) -> RdbReference:
        r"""
        @brief Creates a copy of self
        """

class RdbCell:
    r"""
    @brief A cell inside the report database
    This class represents a cell in the report database. There is not necessarily a 1:1 correspondence of RDB cells and layout database cells. Cells have an ID, a name, optionally a variant name and a set of references which describe at least one example instantiation in some parent cell. The references do not necessarily map to references or cover all references in the layout database.
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def add_reference(self, ref: RdbReference) -> None:
        r"""
        @brief Adds a reference to the references of this cell
        @param ref The reference to add.
        """
    def clear_references(self) -> None:
        r"""
        @brief Removes all references from this cell
        """
    def database(self) -> ReportDatabase:
        r"""
        @brief Gets the database object that category is associated with

        This method has been introduced in version 0.23.
        """
    def each_item(self) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database which are associated with this cell

        This method has been introduced in version 0.23.
        """
    def each_reference(self) -> Iterator[RdbReference]:
        r"""
        @brief Iterates over all references
        """
    def name(self) -> str:
        r"""
        @brief Gets the cell name
        The cell name is an string that identifies the category in the database. Additionally, a cell may carry a variant identifier which is a string that uniquely identifies a cell in the context of it's variants. The "qualified name" contains both the cell name and the variant name. Cell names are also used to identify report database cell's with layout cells. @return The cell name
        """
    def num_items(self) -> int:
        r"""
        @brief Gets the number of items for this cell
        """
    def num_items_visited(self) -> int:
        r"""
        @brief Gets the number of visited items for this cell
        """
    def qname(self) -> str:
        r"""
        @brief Gets the cell's qualified name
        The qualified name is a combination of the cell name and optionally the variant name. It is used to identify the cell by name in a unique way.
        @return The qualified name
        """
    def rdb_id(self) -> int:
        r"""
        @brief Gets the cell ID
        The cell ID is an integer that uniquely identifies the cell. It is used for referring to a cell in \RdbItem for example.
        @return The cell ID
        """
    def variant(self) -> str:
        r"""
        @brief Gets the cell variant name
        A variant name additionally identifies the cell when multiple cells with the same name are present. A variant name is either assigned automatically or set when creating a cell. @return The cell variant name
        """

class RdbCategory:
    r"""
    @brief A category inside the report database
    Every item in the report database is assigned to a category. A category is a DRC rule check for example. Categories can be organized hierarchically, i.e. a category may have sub-categories. Item counts are summarized for categories and items belonging to sub-categories of one category can be browsed together for example. As a general rule, categories not being leaf categories (having child categories) may not have items. 
    """
    description: str
    r"""
    @brief Gets the category description
    @return The description string
    @brief Sets the category description
    @param description The description string
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def database(self) -> ReportDatabase:
        r"""
        @brief Gets the database object that category is associated with

        This method has been introduced in version 0.23.
        """
    def each_item(self) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database which are associated with this category

        This method has been introduced in version 0.23.
        """
    def each_sub_category(self) -> Iterator[RdbCategory]:
        r"""
        @brief Iterates over all sub-categories
        """
    def name(self) -> str:
        r"""
        @brief Gets the category name
        The category name is an string that identifies the category in the context of a parent category or inside the database when it is a top level category. The name is not the path name which is a path to a child category and incorporates all names of parent categories.
        @return The category name
        """
    def num_items(self) -> int:
        r"""
        @brief Gets the number of items in this category
        The number of items includes the items in sub-categories of this category.
        """
    def num_items_visited(self) -> int:
        r"""
        @brief Gets the number of visited items in this category
        The number of items includes the items in sub-categories of this category.
        """
    def parent(self) -> RdbCategory:
        r"""
        @brief Gets the parent category of this category
        @return The parent category or nil if this category is a top-level category
        """
    def path(self) -> str:
        r"""
        @brief Gets the category path
        The category path is the category name for top level categories. For child categories, the path contains the names of all parent categories separated by a dot.
        @return The path for this category
        """
    def rdb_id(self) -> int:
        r"""
        @brief Gets the category ID
        The category ID is an integer that uniquely identifies the category. It is used for referring to a category in \RdbItem for example.
        @return The category ID
        """
    @overload
    def scan_collection(self, cell: RdbCell, trans: db.CplxTrans, edge_pairs: db.EdgePairs, flat: Optional[bool] = ...) -> None:
        r"""
        @brief Turns the given edge pair collection into a hierarchical or flat report database
        This a another flavour of \scan_collection accepting an edge pair collection.

        This method has been introduced in version 0.26.
        """
    @overload
    def scan_collection(self, cell: RdbCell, trans: db.CplxTrans, edges: db.Edges, flat: Optional[bool] = ...) -> None:
        r"""
        @brief Turns the given edge collection into a hierarchical or flat report database
        This a another flavour of \scan_collection accepting an edge collection.

        This method has been introduced in version 0.26.
        """
    @overload
    def scan_collection(self, cell: RdbCell, trans: db.CplxTrans, region: db.Region, flat: Optional[bool] = ...) -> None:
        r"""
        @brief Turns the given region into a hierarchical or flat report database
        The exact behavior depends on the nature of the region. If the region is a hierarchical (original or deep) region and the 'flat' argument is false, this method will produce a hierarchical report database in the given category. The 'cell_id' parameter is ignored in this case. Sample references will be produced to supply minimal instantiation information.

        If the region is a flat one or the 'flat' argument is true, the region's polygons will be produced as report database items in this category and in the cell given by 'cell_id'.

        The transformation argument needs to supply the dbu-to-micron transformation.

        This method has been introduced in version 0.26.
        """
    @overload
    def scan_layer(self, layout: db.Layout, layer: int) -> None:
        r"""
        @brief Scans a layer from a layout into this category
        Creates RDB items for each polygon or edge shape read from the each cell in the layout on the given layer and puts them into this category.
        New cells will be generated for every cell encountered in the layout.
        Other settings like database unit, description, top cell etc. are not made in the RDB.

        This method has been introduced in version 0.23.
        """
    @overload
    def scan_layer(self, layout: db.Layout, layer: int, cell: db.Cell) -> None:
        r"""
        @brief Scans a layer from a layout into this category, starting with a given cell
        Creates RDB items for each polygon or edge shape read from the cell and it's children in the layout on the given layer and puts them into this category.
        New cells will be generated when required.
        Other settings like database unit, description, top cell etc. are not made in the RDB.

        This method has been introduced in version 0.23.
        """
    @overload
    def scan_layer(self, layout: db.Layout, layer: int, cell: db.Cell, levels: int) -> None:
        r"""
        @brief Scans a layer from a layout into this category, starting with a given cell and a depth specification
        Creates RDB items for each polygon or edge shape read from the cell and it's children in the layout on the given layer and puts them into this category.
        New cells will be generated when required.
        "levels" is the number of hierarchy levels to take the child cells from. 0 means to use only "cell" and don't descend, -1 means "all levels".
        Other settings like database unit, description, top cell etc. are not made in the RDB.

        This method has been introduced in version 0.23.
        """
    def scan_shapes(self, iter: db.RecursiveShapeIterator, flat: Optional[bool] = ...) -> None:
        r"""
        @brief Scans the polygon or edge shapes from the shape iterator into the category
        Creates RDB items for each polygon or edge shape read from the iterator and puts them into this category.
        A similar, but lower-level method is \ReportDatabase#create_items with a \RecursiveShapeIterator argument.
        In contrast to \ReportDatabase#create_items, 'scan_shapes' can also produce hierarchical databases if the \flat argument is false. In this case, the hierarchy the recursive shape iterator traverses is copied into the report database using sample references.

        This method has been introduced in version 0.23. The flat mode argument has been added in version 0.26.
        """

class RdbItemValue:
    r"""
    @brief A value object inside the report database
    Value objects are attached to items to provide markers. An arbitrary number of such value objects can be attached to an item.
    Currently, a value can represent a box, a polygon or an edge. Geometrical objects are represented in micron units and are therefore "D" type objects (DPolygon, DEdge and DBox). 
    """
    tag_id: int
    r"""
    @brief Gets the tag ID if the value is a tagged value or 0 if not
    @return The tag ID
    See \tag_id= for details about tagged values.

    Tagged values have been added in version 0.24.
    @brief Sets the tag ID to make the value a tagged value or 0 to reset it
    @param id The tag ID
    To get a tag ID, use \RdbDatabase#user_tag_id (preferred) or \RdbDatabase#tag_id (for internal use).
    Tagged values have been added in version 0.24. Tags can be given to identify a value, for example to attache measurement values to an item. To attach a value for a specific measurement, a tagged value can be used where the tag ID describes the measurement made. In that way, multiple values for different measurements can be attached to an item.

    This variant has been introduced in version 0.24
    """
    def __copy__(self) -> RdbItemValue:
        r"""
        @brief Creates a copy of self
        """
    @overload
    def __init__(self, b: db.DBox) -> None:
        r"""
        @brief Creates a value representing a DBox object
        """
    @overload
    def __init__(self, e: db.DEdge) -> None:
        r"""
        @brief Creates a value representing a DEdge object
        """
    @overload
    def __init__(self, ee: db.DEdgePair) -> None:
        r"""
        @brief Creates a value representing a DEdgePair object
        """
    @overload
    def __init__(self, f: float) -> None:
        r"""
        @brief Creates a value representing a numeric value

        This variant has been introduced in version 0.24
        """
    @overload
    def __init__(self, p: db.DPath) -> None:
        r"""
        @brief Creates a value representing a DPath object

        This method has been introduced in version 0.22.
        """
    @overload
    def __init__(self, p: db.DPolygon) -> None:
        r"""
        @brief Creates a value representing a DPolygon object
        """
    @overload
    def __init__(self, s: str) -> None:
        r"""
        @brief Creates a value representing a string
        """
    @overload
    def __init__(self, t: db.DText) -> None:
        r"""
        @brief Creates a value representing a DText object

        This method has been introduced in version 0.22.
        """
    def __repr__(self) -> str:
        r"""
        @brief Converts a value to a string
        The string can be used by the string constructor to create another object from it.
        @return The string
        """
    def __str__(self) -> str:
        r"""
        @brief Converts a value to a string
        The string can be used by the string constructor to create another object from it.
        @return The string
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: RdbItemValue) -> None:
        r"""
        @brief Assigns another object to self
        """
    def box(self) -> db.DBox:
        r"""
        @brief Gets the box if the value represents one.
        @return The \DBox object or nil
        """
    def dup(self) -> RdbItemValue:
        r"""
        @brief Creates a copy of self
        """
    def edge(self) -> db.DEdge:
        r"""
        @brief Gets the edge if the value represents one.
        @return The \DEdge object or nil
        """
    def edge_pair(self) -> db.DEdgePair:
        r"""
        @brief Gets the edge pair if the value represents one.
        @return The \DEdgePair object or nil
        """
    def float(self) -> float:
        r"""
        @brief Gets the numeric value.
        @return The numeric value or 0
        This method has been introduced in version 0.24.
        """
    def from_s(self, s: str) -> RdbItemValue:
        r"""
        @brief Creates a value object from a string
        The string format is the same than obtained by the to_s method.
        """
    def is_box(self) -> bool:
        r"""
        @brief Returns true if the value object represents a box
        """
    def is_edge(self) -> bool:
        r"""
        @brief Returns true if the value object represents an edge
        """
    def is_edge_pair(self) -> bool:
        r"""
        @brief Returns true if the value object represents an edge pair
        """
    def is_float(self) -> bool:
        r"""
        @brief Returns true if the value object represents a numeric value
        This method has been introduced in version 0.24.
        """
    def is_path(self) -> bool:
        r"""
        @brief Returns true if the value object represents a path

        This method has been introduced in version 0.22.
        """
    def is_polygon(self) -> bool:
        r"""
        @brief Returns true if the value object represents a polygon
        """
    def is_string(self) -> bool:
        r"""
        @brief Returns true if the object represents a string value
        """
    def is_text(self) -> bool:
        r"""
        @brief Returns true if the value object represents a text

        This method has been introduced in version 0.22.
        """
    def path(self) -> db.DPath:
        r"""
        @brief Gets the path if the value represents one.
        @return The \DPath object
        This method has been introduced in version 0.22.
        """
    def polygon(self) -> db.DPolygon:
        r"""
        @brief Gets the polygon if the value represents one.
        @return The \DPolygon object
        """
    def string(self) -> str:
        r"""
        @brief Gets the string representation of the value.
        @return The stringThis method will always deliver a valid string, even if \is_string? is false. The objects stored in the value are converted to a string accordingly.
        """
    def text(self) -> db.DText:
        r"""
        @brief Gets the text if the value represents one.
        @return The \DText object
        This method has been introduced in version 0.22.
        """
    def to_s(self) -> str:
        r"""
        @brief Converts a value to a string
        The string can be used by the string constructor to create another object from it.
        @return The string
        """

class RdbItem:
    r"""
    @brief An item inside the report database
    An item is the basic information entity in the RDB. It is associated with a cell and a category. It can be assigned values which encapsulate other objects such as strings and geometrical objects. In addition, items can be assigned an image (i.e. a screenshot image) and tags which are basically boolean flags that can be defined freely.
    """
    tags_str: str
    r"""
    @brief Returns a string listing all tags of this item
    @return A comma-separated list of tags
    @brief Sets the tags from a string
    @param tags A comma-separated list of tags
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def add_tag(self, tag_id: int) -> None:
        r"""
        @brief Adds a tag with the given id to the item
        Each tag can be added once to the item. The tags of an item thus form a set. If a tag with that ID already exists, this method does nothing.
        """
    @overload
    def add_value(self, value: float) -> None:
        r"""
        @brief Adds a numeric value to the values of this item
        @param value The value to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, value: RdbItemValue) -> None:
        r"""
        @brief Adds a value object to the values of this item
        @param value The value to add.
        """
    @overload
    def add_value(self, value: db.DBox) -> None:
        r"""
        @brief Adds a box object to the values of this item
        @param value The box to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, value: db.DEdge) -> None:
        r"""
        @brief Adds an edge object to the values of this item
        @param value The edge to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, value: db.DEdgePair) -> None:
        r"""
        @brief Adds an edge pair object to the values of this item
        @param value The edge pair to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, value: db.DPolygon) -> None:
        r"""
        @brief Adds a polygon object to the values of this item
        @param value The polygon to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, value: str) -> None:
        r"""
        @brief Adds a string object to the values of this item
        @param value The string to add.
        This method has been introduced in version 0.25 as a convenience method.
        """
    @overload
    def add_value(self, shape: db.Shape, trans: db.CplxTrans) -> None:
        r"""
        @brief Adds a geometrical value object from a shape
        @param value The shape object from which to take the geometrical object.
        @param trans The transformation to apply.

        The transformation can be used to convert database units to micron units.

        This method has been introduced in version 0.25.3.
        """
    def category_id(self) -> int:
        r"""
        @brief Gets the category ID
        Returns the ID of the category that this item is associated with.
        @return The category ID
        """
    def cell_id(self) -> int:
        r"""
        @brief Gets the cell ID
        Returns the ID of the cell that this item is associated with.
        @return The cell ID
        """
    def clear_values(self) -> None:
        r"""
        @brief Removes all values from this item
        """
    def database(self) -> ReportDatabase:
        r"""
        @brief Gets the database object that item is associated with

        This method has been introduced in version 0.23.
        """
    def each_value(self) -> Iterator[RdbItemValue]:
        r"""
        @brief Iterates over all values
        """
    def has_tag(self, tag_id: int) -> bool:
        r"""
        @brief Returns a value indicating whether the item has a tag with the given ID
        @return True, if the item has a tag with the given ID
        """
    def is_visited(self) -> bool:
        r"""
        @brief Gets a value indicating whether the item was already visited
        @return True, if the item has been visited already
        """
    def remove_tag(self, tag_id: int) -> None:
        r"""
        @brief Remove the tag with the given id from the item
        If a tag with that ID does not exists on this item, this method does nothing.
        """

class ReportDatabase:
    r"""
    @brief The report database object
    A report database is organized around a set of items which are associated with cells and categories. Categories can be organized hierarchically by created sub-categories of other categories. Cells are associated with layout database cells and can come with a example instantiation if the layout database does not allow a unique association of the cells.
    Items in the database can have a variety of attributes: values, tags and an image object. Values are geometrical objects for example. Tags are a set of boolean flags and an image can be attached to an item to provide a screenshot for visualization for example.
    This is the main report database object. The basic use case of this object is to create one inside a \LayoutView and populate it with items, cell and categories or load it from a file. Another use case is to create a standalone ReportDatabase object and use the methods provided to perform queries or to populate it.
    """
    description: str
    r"""
    @brief Gets the databases description
    The description is a general purpose string that is supposed to further describe the database and it's content in a human-readable form.
    @return The description string
    @brief Sets the databases description
    @param desc The description string
    """
    generator: str
    r"""
    @brief Gets the databases generator
    The generator string describes how the database was created, i.e. DRC tool name and tool options.
    In a later version this will allow re-running the tool that created the report.
    @return The generator string
    @brief Sets the generator string
    @param generator The generator string
    """
    original_file: str
    r"""
    @brief Gets the original file name and path
    The original file name is supposed to describe the file from which this report database was generated. @return The original file name and path
    @brief Sets the original file name and path
    @param path The path
    """
    top_cell_name: str
    r"""
    @brief Gets the top cell name
    The top cell name identifies the top cell of the design for which the report was generated. This property must be set to establish a proper hierarchical context for a hierarchical report database. @return The top cell name
    @brief Sets the top cell name string
    @param cell_name The top cell name
    """
    def __init__(self, name: str) -> None:
        r"""
        @brief Creates a report database
        @param name The name of the database
        The name of the database will be used in the user interface to refer to a certain database.
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def category_by_id(self, id: int) -> RdbCategory:
        r"""
        @brief Gets a category by ID
        @return The (const) category object or nil if the ID is not valid
        """
    def category_by_path(self, path: str) -> RdbCategory:
        r"""
        @brief Gets a category by path
        @param path The full path to the category starting from the top level (subcategories separated by dots)
        @return The (const) category object or nil if the name is not valid
        """
    def cell_by_id(self, id: int) -> RdbCell:
        r"""
        @brief Returns the cell for a given ID
        @param id The ID of the cell
        @return The cell object or nil if no cell with that ID exists
        """
    def cell_by_qname(self, qname: str) -> RdbCell:
        r"""
        @brief Returns the cell for a given qualified name
        @param qname The qualified name of the cell (name plus variant name optionally)
        @return The cell object or nil if no such cell exists
        """
    @overload
    def create_category(self, name: str) -> RdbCategory:
        r"""
        @brief Creates a new top level category
        @param name The name of the category
        """
    @overload
    def create_category(self, parent: RdbCategory, name: str) -> RdbCategory:
        r"""
        @brief Creates a new sub-category
        @param parent The category under which the category should be created
        @param name The name of the category
        """
    @overload
    def create_cell(self, name: str) -> RdbCell:
        r"""
        @brief Creates a new cell
        @param name The name of the cell
        """
    @overload
    def create_cell(self, name: str, variant: str) -> RdbCell:
        r"""
        @brief Creates a new cell, potentially as a variant for a cell with the same name
        @param name The name of the cell
        @param variant The variant name of the cell
        """
    @overload
    def create_item(self, cell: RdbCell, category: RdbCategory) -> RdbItem:
        r"""
        @brief Creates a new item for the given cell/category combination
        @param cell The cell to which the item is associated
        @param category The category to which the item is associated

        This convenience method has been added in version 0.25.
        """
    @overload
    def create_item(self, cell_id: int, category_id: int) -> RdbItem:
        r"""
        @brief Creates a new item for the given cell/category combination
        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated

        A more convenient method that takes cell and category objects instead of ID's is the other version of \create_item.
        """
    @overload
    def create_item(self, cell_id: int, category_id: int, trans: db.CplxTrans, shape: db.Shape) -> None:
        r"""
        @brief Creates a new item from a single shape
        This method produces an item from the given shape.
        It accepts various kind of shapes, such as texts, polygons, boxes and paths and converts them to a corresponding item. The transformation argument can be used to supply the transformation that applies the database unit for example.

        This method has been introduced in version 0.25.3.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param shape The shape to take the geometrical object from
        @param trans The transformation to apply
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, iter: db.RecursiveShapeIterator) -> None:
        r"""
        @brief Creates new items from a shape iterator
        This method takes the shapes from the given iterator and produces items from them.
        It accepts various kind of shapes, such as texts, polygons, boxes and paths and converts them to corresponding items. This method will produce a flat version of the shapes iterated by the shape iterator. A similar method, which is intended for production of polygon or edge error layers and also provides hierarchical database construction is \RdbCategory#scan_shapes.

        This method has been introduced in version 0.25.3.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param iter The iterator (a \RecursiveShapeIterator object) from which to take the items
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, array: Sequence[db.EdgePair]) -> None:
        r"""
        @brief Creates new edge pair items for the given cell/category combination
        For each edge pair a single item will be created. The value of the item will be this edge pair.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param edge_pairs The list of edge_pairs for which the items are created
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, array: Sequence[db.Edge]) -> None:
        r"""
        @brief Creates new edge items for the given cell/category combination
        For each edge a single item will be created. The value of the item will be this edge.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param edges The list of edges for which the items are created
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, array: Sequence[db.Polygon]) -> None:
        r"""
        @brief Creates new polygon items for the given cell/category combination
        For each polygon a single item will be created. The value of the item will be this polygon.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param polygons The list of polygons for which the items are created
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, edge_pairs: db.EdgePairs) -> None:
        r"""
        @brief Creates new edge pair items for the given cell/category combination
        For each edge pair a single item will be created. The value of the item will be this edge pair.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method will also produce a flat version of the edge pairs inside the edge pair collection. \RdbCategory#scan_collection is a similar method which also supports construction of hierarchical databases from deep edge pair collections.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param edges The list of edge pairs (an \EdgePairs object) for which the items are created
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, edges: db.Edges) -> None:
        r"""
        @brief Creates new edge items for the given cell/category combination
        For each edge a single item will be created. The value of the item will be this edge.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method will also produce a flat version of the edges inside the edge collection. \RdbCategory#scan_collection is a similar method which also supports construction of hierarchical databases from deep edge collections.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param edges The list of edges (an \Edges object) for which the items are created
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, region: db.Region) -> None:
        r"""
        @brief Creates new polygon items for the given cell/category combination
        For each polygon in the region a single item will be created. The value of the item will be this polygon.
        A transformation can be supplied which can be used for example to convert the object's dimensions to micron units by scaling by the database unit.

        This method will also produce a flat version of the shapes inside the region. \RdbCategory#scan_collection is a similar method which also supports construction of hierarchical databases from deep regions.

        This method has been introduced in version 0.23.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param trans The transformation to apply
        @param region The region (a \Region object) containing the polygons for which to create items
        """
    @overload
    def create_items(self, cell_id: int, category_id: int, trans: db.CplxTrans, shapes: db.Shapes) -> None:
        r"""
        @brief Creates new items from a shape container
        This method takes the shapes from the given container and produces items from them.
        It accepts various kind of shapes, such as texts, polygons, boxes and paths and converts them to corresponding items. The transformation argument can be used to supply the transformation that applies the database unit for example.

        This method has been introduced in version 0.25.3.

        @param cell_id The ID of the cell to which the item is associated
        @param category_id The ID of the category to which the item is associated
        @param shapes The shape container from which to take the items
        @param trans The transformation to apply
        """
    def each_category(self) -> Iterator[RdbCategory]:
        r"""
        @brief Iterates over all top-level categories
        """
    def each_cell(self) -> Iterator[RdbCell]:
        r"""
        @brief Iterates over all cells
        """
    def each_item(self) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database
        """
    def each_item_per_category(self, category_id: int) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database which are associated with the given category
        @param category_id The ID of the category for which all associated items should be retrieved
        """
    def each_item_per_cell(self, cell_id: int) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database which are associated with the given cell
        @param cell_id The ID of the cell for which all associated items should be retrieved
        """
    def each_item_per_cell_and_category(self, cell_id: int, category_id: int) -> Iterator[RdbItem]:
        r"""
        @brief Iterates over all items inside the database which are associated with the given cell and category
        @param cell_id The ID of the cell for which all associated items should be retrieved
        @param category_id The ID of the category for which all associated items should be retrieved
        """
    def filename(self) -> str:
        r"""
        @brief Gets the file name and path where the report database is stored
        This property is set when a database is saved or loaded. It cannot be set manually.
        @return The file name and path
        """
    def is_modified(self) -> bool:
        r"""
        @brief Returns a value indicating whether the database has been modified
        """
    def load(self, filename: str) -> None:
        r"""
        @brief Loads the database from the given file
        @param filename The file from which to load the database
        The reader recognizes the format automatically and will choose the appropriate decoder. 'gzip' compressed files are uncompressed automatically.
        """
    def name(self) -> str:
        r"""
        @brief Gets the database name
        The name of the database is supposed to identify the database within a layout view context. The name is modified to be unique when a database is entered into a layout view. @return The database name
        """
    @overload
    def num_items(self) -> int:
        r"""
        @brief Returns the number of items inside the database
        @return The total number of items
        """
    @overload
    def num_items(self, cell_id: int, category_id: int) -> int:
        r"""
        @brief Returns the number of items inside the database for a given cell/category combination
        @param cell_id The ID of the cell for which to retrieve the number
        @param category_id The ID of the category for which to retrieve the number
        @return The total number of items for the given cell and the given category
        """
    @overload
    def num_items_visited(self) -> int:
        r"""
        @brief Returns the number of items already visited inside the database
        @return The total number of items already visited
        """
    @overload
    def num_items_visited(self, cell_id: int, category_id: int) -> int:
        r"""
        @brief Returns the number of items visited already for a given cell/category combination
        @param cell_id The ID of the cell for which to retrieve the number
        @param category_id The ID of the category for which to retrieve the number
        @return The total number of items visited for the given cell and the given category
        """
    def reset_modified(self) -> None:
        r"""
        @brief Reset the modified flag
        """
    def save(self, filename: str) -> None:
        r"""
        @brief Saves the database to the given file
        @param filename The file to which to save the database
        The database is always saved in KLayout's XML-based format.
        """
    def set_item_visited(self, item: RdbItem, visited: bool) -> None:
        r"""
        @brief Modifies the visited state of an item
        @param item The item to modify
        @param visited True to set the item to visited state, false otherwise
        """
    def set_tag_description(self, tag_id: int, description: str) -> None:
        r"""
        @brief Sets the tag description for the given tag ID
        @param tag_id The ID of the tag
        @param description The description string
        See \tag_id for a details about tags.
        """
    def tag_description(self, tag_id: int) -> str:
        r"""
        @brief Gets the tag description for the given tag ID
        @param tag_id The ID of the tag
        @return The description string
        See \tag_id for a details about tags.
        """
    def tag_id(self, name: str) -> int:
        r"""
        @brief Gets the tag ID for a given tag name
        @param name The tag name
        @return The corresponding tag ID
        Tags are used to tag items in the database and to specify tagged (named) values. This method will always succeed and the tag will be created if it does not exist yet. Tags are basically names. There are user tags (for free assignment) and system tags which are used within the system. Both are separated to avoid name clashes.

        \tag_id handles system tags while \user_tag_id handles user tags.
        """
    def tag_name(self, tag_id: int) -> str:
        r"""
        @brief Gets the tag name for the given tag ID
        @param tag_id The ID of the tag
        @return The name of the tag
        See \tag_id for a details about tags.

        This method has been introduced in version 0.24.10.
        """
    def user_tag_id(self, name: str) -> int:
        r"""
        @brief Gets the tag ID for a given user tag name
        @param name The user tag name
        @return The corresponding tag ID
        This method will always succeed and the tag will be created if it does not exist yet. See \tag_id for a details about tags.

        This method has been added in version 0.24.
        """
    def variants(self, name: str) -> List[int]:
        r"""
        @brief Gets the variants for a given cell name
        @param name The basic name of the cell
        @return An array of ID's representing cells that are variants for the given base name
        """




from typing import Any, ClassVar, Dict, Sequence, List, Iterator, Optional
from typing import overload
class EmptyClass:
    r"""
    """
    def __copy__(self) -> EmptyClass:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: EmptyClass) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> EmptyClass:
        r"""
        @brief Creates a copy of self
        """

class Value:
    r"""
    @brief Encapsulates a value (preferably a plain data type) in an object
    This class is provided to 'box' a value (encapsulate the value in an object). This class is required to interface to pointer or reference types in a method call. By using that class, the method can alter the value and thus implement 'out parameter' semantics. The value may be 'nil' which acts as a null pointer in pointer type arguments.
    This class has been introduced in version 0.22.
    """
    value: Any
    r"""
    @brief Gets the actual value.
    @brief Set the actual value.
    """
    def __copy__(self) -> Value:
        r"""
        @brief Creates a copy of self
        """
    @overload
    def __init__(self) -> None:
        r"""
        @brief Constructs a nil object.
        """
    @overload
    def __init__(self, value: Any) -> None:
        r"""
        @brief Constructs a non-nil object with the given value.
        This constructor has been introduced in version 0.22.
        """
    def __repr__(self) -> str:
        r"""
        @brief Convert this object to a string
        """
    def __str__(self) -> str:
        r"""
        @brief Convert this object to a string
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: Value) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> Value:
        r"""
        @brief Creates a copy of self
        """
    def to_s(self) -> str:
        r"""
        @brief Convert this object to a string
        """

class Interpreter:
    r"""
    @brief A generalization of script interpreters
    The main purpose of this class is to provide cross-language call options. Using the Python interpreter, it is possible to execute Python code from Ruby for example.

    The following example shows how to use the interpreter class to execute Python code from Ruby and how to pass values from Ruby to Python and back using the \Value wrapper object:

    @code
    pya = RBA::Interpreter::python_interpreter
    out_param = RBA::Value::new(17)
    pya.define_variable("out_param", out_param)
    pya.eval_string(<<END)
    print("This is Python now!")
    out_param.value = out_param.value + 25
    END
    puts out_param.value  # gives '42'@/code

    This class was introduced in version 0.27.5.
    """
    @classmethod
    def python_interpreter(cls) -> Interpreter:
        r"""
        @brief Gets the instance of the Python interpreter
        """
    @classmethod
    def ruby_interpreter(cls) -> Interpreter:
        r"""
        @brief Gets the instance of the Ruby interpreter
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def define_variable(self, name: str, value: Any) -> None:
        r"""
        @brief Defines a (global) variable with the given name and value
        You can use the \Value class to provide 'out' or 'inout' parameters which can be modified by code executed inside the interpreter and read back by the caller.
        """
    def eval_expr(self, string: str, filename: Optional[str] = ..., line: Optional[int] = ...) -> Any:
        r"""
        @brief Executes the expression inside the given string and returns the result value
        Use 'filename' and 'line' to indicate the original source for the error messages.
        """
    def eval_string(self, string: str, filename: Optional[str] = ..., line: Optional[int] = ...) -> None:
        r"""
        @brief Executes the code inside the given string
        Use 'filename' and 'line' to indicate the original source for the error messages.
        """
    def load_file(self, path: str) -> None:
        r"""
        @brief Loads the given file into the interpreter
        This will execute the code inside the file.
        """

class ArgType:
    r"""
    @hide
    """
    TypeBool: ClassVar[int]
    r"""
    """
    TypeChar: ClassVar[int]
    r"""
    """
    TypeDouble: ClassVar[int]
    r"""
    """
    TypeFloat: ClassVar[int]
    r"""
    """
    TypeInt: ClassVar[int]
    r"""
    """
    TypeLong: ClassVar[int]
    r"""
    """
    TypeLongLong: ClassVar[int]
    r"""
    """
    TypeMap: ClassVar[int]
    r"""
    """
    TypeObject: ClassVar[int]
    r"""
    """
    TypeSChar: ClassVar[int]
    r"""
    """
    TypeShort: ClassVar[int]
    r"""
    """
    TypeString: ClassVar[int]
    r"""
    """
    TypeUChar: ClassVar[int]
    r"""
    """
    TypeUInt: ClassVar[int]
    r"""
    """
    TypeULong: ClassVar[int]
    r"""
    """
    TypeULongLong: ClassVar[int]
    r"""
    """
    TypeUShort: ClassVar[int]
    r"""
    """
    TypeVar: ClassVar[int]
    r"""
    """
    TypeVector: ClassVar[int]
    r"""
    """
    TypeVoid: ClassVar[int]
    r"""
    """
    TypeVoidPtr: ClassVar[int]
    r"""
    """
    def __copy__(self) -> ArgType:
        r"""
        @brief Creates a copy of self
        """
    def __eq__(self, arg0: object) -> bool:
        r"""
        @brief Equality of two types
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def __ne__(self, arg0: object) -> bool:
        r"""
        @brief Inequality of two types
        """
    def __repr__(self) -> str:
        r"""
        @brief Convert to a string
        """
    def __str__(self) -> str:
        r"""
        @brief Convert to a string
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: ArgType) -> None:
        r"""
        @brief Assigns another object to self
        """
    def cls(self) -> Class:
        r"""
        @brief Specifies the class for t_object.. types
        """
    def default(self) -> Any:
        r"""
        @brief Returns the default value or nil is there is no default value
        Applies to arguments only. This method has been introduced in version 0.24.
        """
    def dup(self) -> ArgType:
        r"""
        @brief Creates a copy of self
        """
    def has_default(self) -> bool:
        r"""
        @brief Returns true, if a default value is specified for this argument
        Applies to arguments only. This method has been introduced in version 0.24.
        """
    def inner(self) -> ArgType:
        r"""
        @brief Returns the inner ArgType object (i.e. value of a vector/map)
        Starting with version 0.22, this method replaces the is_vector method.
        """
    def inner_k(self) -> ArgType:
        r"""
        @brief Returns the inner ArgType object (i.e. key of a map)
        This method has been introduced in version 0.27.
        """
    def is_cptr(self) -> bool:
        r"""
        @brief True, if the type is a const pointer to the given type
        This property indicates that the argument is a const pointer (in C++: 'const X *').
        """
    def is_cref(self) -> bool:
        r"""
        @brief True, if the type is a const reference to the given type
        This property indicates that the argument is a const reference (in C++: 'const X &').
        """
    def is_iter(self) -> bool:
        r"""
        @brief (Return value only) True, if the return value is an iterator rendering the given type
        """
    def is_ptr(self) -> bool:
        r"""
        @brief True, if the type is a non-const pointer to the given type
        This property indicates that the argument is a non-const pointer (in C++: 'X *').
        """
    def is_ref(self) -> bool:
        r"""
        @brief True, if the type is a reference to the given type
        Starting with version 0.22 there are more methods that describe the type of reference and is_ref? only applies to non-const reference (in C++: 'X &').
        """
    def name(self) -> str:
        r"""
        @brief Returns the name for this argument or an empty string if the argument is not named
        Applies to arguments only. This method has been introduced in version 0.24.
        """
    def pass_obj(self) -> bool:
        r"""
        @brief True, if the ownership over an object represented by this type is passed to the receiver
        In case of the return type, a value of true indicates, that the object is a freshly created one and the receiver has to take ownership of the object.

        This method has been introduced in version 0.24.
        """
    def to_s(self) -> str:
        r"""
        @brief Convert to a string
        """
    def type(self) -> int:
        r"""
        @brief Return the basic type (see t_.. constants)
        """

class MethodOverload:
    r"""
    @hide
    """
    def __copy__(self) -> MethodOverload:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: MethodOverload) -> None:
        r"""
        @brief Assigns another object to self
        """
    def deprecated(self) -> bool:
        r"""
        @brief A value indicating that this overload is deprecated
        """
    def dup(self) -> MethodOverload:
        r"""
        @brief Creates a copy of self
        """
    def is_getter(self) -> bool:
        r"""
        @brief A value indicating that this overload is a property getter
        """
    def is_predicate(self) -> bool:
        r"""
        @brief A value indicating that this overload is a predicate
        """
    def is_setter(self) -> bool:
        r"""
        @brief A value indicating that this overload is a property setter
        """
    def name(self) -> str:
        r"""
        @brief The name of this overload
        This is the raw, unadorned name. I.e. no question mark suffix for predicates, no equal character suffix for setters etc.
        """

class Method:
    r"""
    @hide
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def accepts_num_args(self, arg0: int) -> bool:
        r"""
        @brief True, if this method is compatible with the given number of arguments

        This method has been introduced in version 0.24.
        """
    def doc(self) -> str:
        r"""
        @brief The documentation string for this method
        """
    def each_argument(self) -> Iterator[ArgType]:
        r"""
        @brief Iterate over all arguments of this method
        """
    def each_overload(self) -> Iterator[MethodOverload]:
        r"""
        @brief This iterator delivers the synonyms (overloads).

        This method has been introduced in version 0.24.
        """
    def is_const(self) -> bool:
        r"""
        @brief True, if this method does not alter the object
        """
    def is_constructor(self) -> bool:
        r"""
        @brief True, if this method is a constructor
        Static methods that return new objects are constructors.
        This method has been introduced in version 0.25.
        """
    def is_protected(self) -> bool:
        r"""
        @brief True, if this method is protected

        This method has been introduced in version 0.24.
        """
    def is_signal(self) -> bool:
        r"""
        @brief True, if this method is a signal

        Signals replace events for version 0.25. is_event? is no longer available.
        """
    def is_static(self) -> bool:
        r"""
        @brief True, if this method is static (a class method)
        """
    def name(self) -> str:
        r"""
        @brief The name string of the method
        A method may have multiple names (aliases). The name string delivers all of them in a combined way.

        The names are separated by pipe characters (|). A trailing star (*) indicates that the method is protected.

        Names may be prefixed by a colon (:) to indicate a property getter. This colon does not appear in the method name.

        A hash prefix indicates that a specific alias is deprecated.

        Names may be suffixed by a question mark (?) to indicate a predicate or a equal character (=) to indicate a property setter. Depending on the preferences of the language, these characters may appear in the method names of not - in Python they don't, in Ruby they will be part of the method name.

        The backslash character is used inside the names to escape these special characters.

        The preferred method of deriving the overload is to iterate then using \each_overload.
        """
    def primary_name(self) -> str:
        r"""
        @brief The primary name of the method
        The primary name is the first name of a sequence of aliases.

        This method has been introduced in version 0.24.
        """
    def ret_type(self) -> ArgType:
        r"""
        @brief The return type of this method
        """

class Class:
    r"""
    @hide
    """
    @classmethod
    def each_class(cls) -> Iterator[Class]:
        r"""
        @brief Iterate over all classes
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def base(self) -> Class:
        r"""
        @brief The base class or nil if the class does not have a base class

        This method has been introduced in version 0.22.
        """
    def can_copy(self) -> bool:
        r"""
        @brief True if the class offers assignment
        """
    def can_destroy(self) -> bool:
        r"""
        @brief True if the class offers a destroy method

        This method has been introduced in version 0.22.
        """
    def doc(self) -> str:
        r"""
        @brief The documentation string for this class
        """
    def each_child_class(self) -> Iterator[Class]:
        r"""
        @brief Iterate over all child classes defined within this class
        """
    def each_method(self) -> Iterator[Method]:
        r"""
        @brief Iterate over all methods of this class
        """
    def module(self) -> str:
        r"""
        @brief The name of module where the class lives
        """
    def name(self) -> str:
        r"""
        @brief The name of the class
        """
    def parent(self) -> Class:
        r"""
        @brief The parent of the class
        """

class Logger:
    r"""
    @brief A logger

    The logger outputs messages to the log channels. If the log viewer is open, the log messages will be shown in the logger view. Otherwise they will be printed to the terminal on Linux for example.

    A code example:

    @code
    RBA::Logger::error("An error message")
    RBA::Logger::warn("A warning")
    @/code

    This class has been introduced in version 0.23.
    """
    verbosity: int
    r"""
    @brief Returns the verbosity level

    The verbosity level is defined by the application (see -d command line option for example). Level 0 is silent, levels 10, 20, 30 etc. denote levels with increasing verbosity. 11, 21, 31 .. are sublevels which also enable timing logs in addition to messages.@brief Sets the verbosity level for the application

    See \verbosity for a definition of the verbosity levels. Please note that this method changes the verbosity level for the whole application.
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def error(self, msg: str) -> None:
        r"""
        @brief Writes the given string to the error channel

        The error channel is formatted as an error (i.e. red in the logger window) and output unconditionally.
        """
    def info(self, msg: str) -> None:
        r"""
        @brief Writes the given string to the info channel

        The info channel is printed as neutral messages unconditionally.
        """
    def log(self, msg: str) -> None:
        r"""
        @brief Writes the given string to the log channel

        Log messages are printed as neutral messages and are output only if the verbosity is above 0.
        """
    def warn(self, msg: str) -> None:
        r"""
        @brief Writes the given string to the warning channel

        The warning channel is formatted as a warning (i.e. blue in the logger window) and output unconditionally.
        """

class Timer:
    r"""
    @brief A timer (stop watch)

    The timer provides a way to measure CPU time. It provides two basic methods: start and stop. After it has been started and stopped again, the time can be retrieved using the user and sys attributes, i.e.:

    @code
    t = RBA::Timer::new
    t.start
    # ... do something
    t.stop
    puts "it took #{t.sys} seconds (kernel), #{t.user} seconds (user) on the CPU"
    @/code

    The time is reported in seconds.

    This class has been introduced in version 0.23.
    """
    @classmethod
    def memory_size(cls) -> int:
        r"""
        @brief Gets the current memory usage of the process in Bytes

        This method has been introduced in version 0.27.
        """
    def __copy__(self) -> Timer:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def __repr__(self) -> str:
        r"""
        @brief Produces a string with the currently elapsed times
        """
    def __str__(self) -> str:
        r"""
        @brief Produces a string with the currently elapsed times
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: Timer) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> Timer:
        r"""
        @brief Creates a copy of self
        """
    def start(self) -> None:
        r"""
        @brief Starts the timer
        """
    def stop(self) -> None:
        r"""
        @brief Stops the timer
        """
    def sys(self) -> float:
        r"""
        @brief Returns the elapsed CPU time in kernel mode from start to stop in seconds
        """
    def to_s(self) -> str:
        r"""
        @brief Produces a string with the currently elapsed times
        """
    def user(self) -> float:
        r"""
        @brief Returns the elapsed CPU time in user mode from start to stop in seconds
        """
    def wall(self) -> float:
        r"""
        @brief Returns the elapsed real time from start to stop in seconds
        This method has been introduced in version 0.26.
        """

class Progress:
    r"""
    @brief A progress reporter

    This is the base class for all progress reporter objects. Progress reporter objects are used to report the progress of some operation and to allow aborting an operation. Progress reporter objects must be triggered periodically, i.e. a value must be set. On the display side, a progress bar usually is used to represent the progress of an operation.

    Actual implementations of the progress reporter class are \RelativeProgress and \AbsoluteProgress.

    This class has been introduced in version 0.23.
    """
    desc: str
    r"""
    @brief Gets the description text of the progress object
    @brief Sets the description text of the progress object
    """
    title: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the title text of the progress object

    Initially the title is equal to the description.
    """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """

class AbstractProgress(Progress):
    r"""
    @brief The abstract progress reporter

    The abstract progress reporter acts as a 'bracket' for a sequence of operations which are connected logically. For example, a DRC script consists of multiple operations. An abstract progress reportert is instantiated during the run time of the DRC script. This way, the application leaves the UI open while the DRC executes and log messages can be collected.

    The abstract progress does not have a value.

    This class has been introduced in version 0.27.
    """
    def __init__(self, desc: str) -> None:
        r"""
        @brief Creates an abstract progress reporter with the given description
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """

class RelativeProgress(Progress):
    r"""
    @brief A progress reporter counting progress in relative units

    A relative progress reporter counts from 0 to some maximum value representing 0 to 100 percent completion of a task. The progress can be configured to have a description text, a title and a format.
    The "inc" method increments the value, the "set" or "value=" methods set the value to a specific value.

    While one of these three methods is called, they will run the event loop in regular intervals. That makes the application respond to mouse clicks, specifically the Cancel button on the progress bar. If that button is clicked, an exception will be raised by these methods.

    The progress object must be destroyed explicitly in order to remove the progress status bar.

    A code example:

    @code
    p = RBA::RelativeProgress::new("test", 10000000)
    begin
      10000000.times { p.inc }
    ensure
      p.destroy
    end
    @/code

    This class has been introduced in version 0.23.
    """
    format: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief sets the output format (sprintf notation) for the progress text
    """
    value: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the progress value
    """
    @overload
    def __init__(self, desc: str, max_value: int) -> None:
        r"""
        @brief Creates a relative progress reporter with the given description and maximum value

        The reported progress will be 0 to 100% for values between 0 and the maximum value.
        The values are always integers. Double values cannot be used property.
        """
    @overload
    def __init__(self, desc: str, max_value: int, yield_interval: int) -> None:
        r"""
        @brief Creates a relative progress reporter with the given description and maximum value

        The reported progress will be 0 to 100% for values between 0 and the maximum value.
        The values are always integers. Double values cannot be used property.

        The yield interval specifies, how often the event loop will be triggered. When the yield interval is 10 for example, the event loop will be executed every tenth call of \inc or \set.
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def inc(self) -> RelativeProgress:
        r"""
        @brief Increments the progress value
        """
    def set(self, value: int, force_yield: bool) -> None:
        r"""
        @brief Sets the progress value

        This method is equivalent to \value=, but it allows forcing the event loop to be triggered.
        If "force_yield" is true, the event loop will be triggered always, irregardless of the yield interval specified in the constructor.
        """

class AbsoluteProgress(Progress):
    r"""
    @brief A progress reporter counting progress in absolute units

    An absolute progress reporter counts from 0 upwards without a known limit. A unit value is used to convert the value to a bar value. One unit corresponds to 1% on the bar.
    For formatted output, a format string can be specified as well as a unit value by which the current value is divided before it is formatted.

    The progress can be configured to have a description text, a title and a format.
    The "inc" method increments the value, the "set" or "value=" methods set the value to a specific value.

    While one of these three methods is called, they will run the event loop in regular intervals. That makes the application respond to mouse clicks, specifically the Cancel button on the progress bar. If that button is clicked, an exception will be raised by these methods.

    The progress object must be destroyed explicitly in order to remove the progress status bar.

    The following sample code creates a progress bar which displays the current count as "Megabytes".
    For the progress bar, one percent corresponds to 16 kByte:

    @code
    p = RBA::AbsoluteProgress::new("test")
    p.format = "%.2f MBytes"
    p.unit = 1024*16
    p.format_unit = 1024*1024
    begin
      10000000.times { p.inc }
    ensure
      p.destroy
    end
    @/code

    This class has been introduced in version 0.23.
    """
    format: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief sets the output format (sprintf notation) for the progress text
    """
    format_unit: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the format unit

    This is the unit used for formatted output.
    The current count is divided by the format unit to render
    the value passed to the format string.
    """
    unit: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the unit

    Specifies the count value corresponding to 1 percent on the progress bar. By default, the current value divided by the unit is used to create the formatted value from the output string. Another attribute is provided (\format_unit=) to specify a separate unit for that purpose.
    """
    value: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the progress value
    """
    @overload
    def __init__(self, desc: str) -> None:
        r"""
        @brief Creates an absolute progress reporter with the given description
        """
    @overload
    def __init__(self, desc: str, yield_interval: int) -> None:
        r"""
        @brief Creates an absolute progress reporter with the given description

        The yield interval specifies, how often the event loop will be triggered. When the yield interval is 10 for example, the event loop will be executed every tenth call of \inc or \set.
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def inc(self) -> AbsoluteProgress:
        r"""
        @brief Increments the progress value
        """
    def set(self, value: int, force_yield: bool) -> None:
        r"""
        @brief Sets the progress value

        This method is equivalent to \value=, but it allows forcing the event loop to be triggered.
        If "force_yield" is true, the event loop will be triggered always, irregardless of the yield interval specified in the constructor.
        """

class ExpressionContext:
    r"""
    @brief Represents the context of an expression evaluation

    The context provides a variable namespace for the expression evaluation.

    This class has been introduced in version 0.26 when \Expression was separated into the execution and context part.
    """
    def __copy__(self) -> ExpressionContext:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: ExpressionContext) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> ExpressionContext:
        r"""
        @brief Creates a copy of self
        """
    def eval(self, expr: str) -> Any:
        r"""
        @brief Compiles and evaluates the given expression in this context
        This method has been introduced in version 0.26.
        """
    def global_var(self, name: str, value: Any) -> None:
        r"""
        @brief Defines a global variable with the given name and value
        """
    def var(self, name: str, value: Any) -> None:
        r"""
        @brief Defines a variable with the given name and value
        """

class Expression(ExpressionContext):
    r"""
    @brief Evaluation of Expressions

    This class allows evaluation of expressions. Expressions are used in many places throughout KLayout and provide computation features for various applications. Having a script language, there is no real use for expressions inside a script client. This class is provided mainly for testing purposes.

    An expression is 'compiled' into an Expression object and can be evaluated multiple times.

    This class has been introduced in version 0.25. In version 0.26 it was separated into execution and context.
    """
    text: None
    r"""
    WARNING: This variable can only be set, not retrieved.
    @brief Sets the given text as the expression.
    """
    @overload
    def __init__(self, expr: str) -> None:
        r"""
        @brief Creates an expression evaluator
        """
    @overload
    def __init__(self, expr: str, variables: Dict[str, Any]) -> None:
        r"""
        @brief Creates an expression evaluator
        This version of the constructor takes a hash of variables available to the expressions.
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    @overload
    def eval(self) -> Any:
        r"""
        @brief Evaluates the current expression and returns the result
        """
    @overload
    def eval(self, expr: str) -> Any:
        r"""
        @brief A convience function to evaluate the given expression and directly return the result
        This is a static method that does not require instantiation of the expression object first.
        """

class GlobPattern:
    r"""
    @brief A glob pattern matcher
    This class is provided to make KLayout's glob pattern matching available to scripts too. The intention is to provide an implementation which is compatible with KLayout's pattern syntax.

    This class has been introduced in version 0.26.
    """
    case_sensitive: bool
    r"""
    @brief Gets a value indicating whether the glob pattern match is case sensitive.@brief Sets a value indicating whether the glob pattern match is case sensitive.
    """
    head_match: bool
    r"""
    @brief Gets a value indicating whether trailing characters are allowed.
    @brief Sets a value indicating whether trailing characters are allowed.
    If this predicate is false, the glob pattern needs to match the full subject string. If true, the match function will ignore trailing characters and return true if the front part of the subject string matches.
    """
    def __copy__(self) -> GlobPattern:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self, pattern: str) -> None:
        r"""
        @brief Creates a new glob pattern match object
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: GlobPattern) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> GlobPattern:
        r"""
        @brief Creates a copy of self
        """
    def match(self, subject: str) -> Any:
        r"""
        @brief Matches the subject string against the pattern.
        Returns nil if the subject string does not match the pattern. Otherwise returns a list with the substrings captured in round brackets.
        """

class ExecutableBase:
    r"""
    @hide
    @alias Executable
    """
    def __copy__(self) -> ExecutableBase:
        r"""
        @brief Creates a copy of self
        """
    def __init__(self) -> None:
        r"""
        @brief Creates a new object of this class
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def assign(self, other: ExecutableBase) -> None:
        r"""
        @brief Assigns another object to self
        """
    def dup(self) -> ExecutableBase:
        r"""
        @brief Creates a copy of self
        """

class Executable(ExecutableBase):
    r"""
    @brief A generic executable object
    This object is a delegate for implementing the actual function of some generic executable function. In addition to the plain execution, if offers a post-mortem cleanup callback which is always executed, even if execute's implementation is cancelled in the debugger.

    Parameters are kept as a generic key/value map.

    This class has been introduced in version 0.27.
    """
    def _assign(self, other: ExecutableBase) -> None:
        r"""
        @brief Assigns another object to self
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _dup(self) -> Executable:
        r"""
        @brief Creates a copy of self
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def cleanup(self) -> None:
        r"""
        @brief Reimplement this method to provide post-mortem cleanup functionality.
        This method is always called after execute terminated.
        """
    def execute(self) -> Any:
        r"""
        @brief Reimplement this method to provide the functionality of the executable.
        This method is supposed to execute the operation with the given parameters and return the desired output.
        """

class Recipe:
    r"""
    @brief A facility for providing reproducible recipes
    The idea of this facility is to provide a service by which an object
    can be reproduced in a parametrized way. The intended use case is a 
    DRC report for example, where the DRC script is the generator.

    In this use case, the DRC engine will register a recipe. It will 
    put the serialized version of the recipe into the DRC report. If the 
    user requests a re-run of the DRC, the recipe will be called and 
    the implementation is supposed to deliver a new database.

    To register a recipe, reimplement the Recipe class and create an
    instance. To serialize a recipe, use "generator", to execute the
    recipe, use "make".

    Parameters are kept as a generic key/value map.

    This class has been introduced in version 0.26.
    """
    def __init__(self, name: str, description: Optional[str] = ...) -> None:
        r"""
        @brief Creates a new recipe object with the given name and (optional) description
        """
    def _create(self) -> None:
        r"""
        @brief Ensures the C++ object is created
        Use this method to ensure the C++ object is created, for example to ensure that resources are allocated. Usually C++ objects are created on demand and not necessarily when the script object is created.
        """
    def _destroy(self) -> None:
        r"""
        @brief Explicitly destroys the object
        Explicitly destroys the object on C++ side if it was owned by the script interpreter. Subsequent access to this object will throw an exception.
        If the object is not owned by the script, this method will do nothing.
        """
    def _destroyed(self) -> bool:
        r"""
        @brief Returns a value indicating whether the object was already destroyed
        This method returns true, if the object was destroyed, either explicitly or by the C++ side.
        The latter may happen, if the object is owned by a C++ object which got destroyed itself.
        """
    def _is_const_object(self) -> bool:
        r"""
        @brief Returns a value indicating whether the reference is a const reference
        This method returns true, if self is a const reference.
        In that case, only const methods may be called on self.
        """
    def _manage(self) -> None:
        r"""
        @brief Marks the object as managed by the script side.
        After calling this method on an object, the script side will be responsible for the management of the object. This method may be called if an object is returned from a C++ function and the object is known not to be owned by any C++ instance. If necessary, the script side may delete the object if the script's reference is no longer required.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def _unmanage(self) -> None:
        r"""
        @brief Marks the object as no longer owned by the script side.
        Calling this method will make this object no longer owned by the script's memory management. Instead, the object must be managed in some other way. Usually this method may be called if it is known that some C++ object holds and manages this object. Technically speaking, this method will turn the script's reference into a weak reference. After the script engine decides to delete the reference, the object itself will still exist. If the object is not managed otherwise, memory leaks will occur.

        Usually it's not required to call this method. It has been introduced in version 0.24.
        """
    def description(self) -> str:
        r"""
        @brief Gets the description of the recipe.
        """
    def executable(self, params: Dict[str, Any]) -> ExecutableBase:
        r"""
        @brief Reimplement this method to provide an executable object for the actual implementation.
        The reasoning behind this architecture is to supply a cleanup callback. This is useful when the actual function is executed as a script and the script terminates in the debugger. The cleanup callback allows implementing any kind of post-mortem action despite being cancelled in the debugger.

        This method has been introduced in version 0.27 and replaces 'execute'.
        """
    def generator(self, params: Dict[str, Any]) -> str:
        r"""
        @brief Delivers the generator string from the given parameters.
        The generator string can be used with \make to re-run the recipe.
        """
    def make(self, generator: str, add_params: Optional[Dict[str, Any]] = ...) -> Any:
        r"""
        @brief Executes the recipe given by the generator string.
        The generator string is the one delivered with \generator.
        Additional parameters can be passed in "add_params". They have lower priority than the parameters kept inside the generator string.
        """
    def name(self) -> str:
        r"""
        @brief Gets the name of the recipe.
        """




# import all packages from klayout, such as klayout.db and klayout.tl
# WARNING: doing it manually until it becomes impractical
# TODO: We need a specification document explaining what should go into pya

from klayout.db import *  # noqa
from klayout.lib import *  # noqa
from klayout.tl import *  # noqa
from klayout.rdb import *  # noqa



"""

KLayout standalone Python module setup script


    Copyright (C) 2006-2022 Matthias Koefferlein

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


This script provides distutils-based deployment for KLayout's
standalone Python modules.

The standalone libraries are basically extension modules.
Build requirements are:
 * curl library
 * expat library

The main challenge is to map KLayout's shared object architecture.
The structure consists of the Python extension libraries and a bunch
of libraries providing the actual implementation part.
The Python extension libraries reference the implementation libraries.
The extension libraries are prefixed with an underscore.

For example:

    tl (Python) -> _tl, _gsi

There is a specific issue with the building of the extension
libraries: distutils only allows building of shared objects
as extension libraries and does not provide a intrinsic feature
for linking other libraries into the extension library.

Hence we need to add the dependencies through "extra_objects".

On Linux there is a specific issue: The shared objects produced
for extension libraries are not matching the Linux name conventions.
So we have to add them as .so link objects including the path.
If we do so, the runtime loading will look them up by their path
and won't find them. So we need to take away the path with
"-Wl,-soname" on Linux (see Config.link_args).
"""

from typing import List
from setuptools import setup, Distribution, find_packages
from setuptools.extension import Extension, Library
import glob
import os
import re
import sys
import platform
from distutils.errors import CompileError
import distutils.command.build_ext
import setuptools.command.build_ext
import multiprocessing

# for Jenkins we do not want to be greedy
multicore = os.getenv("KLAYOUT_SETUP_MULTICORE")
if multicore:
    N_cores = int(multicore)
else:
    N_cores = multiprocessing.cpu_count()


# monkey-patch for parallel compilation
# from https://stackoverflow.com/questions/11013851/speeding-up-build-process-with-distutils
def parallelCCompile(
    self,
    sources,
    output_dir=None,
    macros=None,
    include_dirs=None,
    debug=0,
    extra_preargs=None,
    extra_postargs=None,
    depends=None,
):
    # those lines are copied from distutils.ccompiler.CCompiler directly
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(
        output_dir, macros, include_dirs, sources, depends, extra_postargs
    )
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
    # parallel code

    N = min(N_cores, len(objects) // 2, 8)  # number of parallel compilations
    N = max(N, 1)
    print("Compiling with", N, "threads.")
    import multiprocessing.pool

    def _single_compile(obj):
        try:
            src, ext = build[obj]
        except KeyError:
            return
        n_tries = 2
        while n_tries > 0:
            try:
                print("Building", obj)
                self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
            except CompileError:
                n_tries -= 1
                print("Building", obj, "has failed. Trying again.")
            else:
                break

    # convert to list, imap is evaluated on-demand
    list(multiprocessing.pool.ThreadPool(N).map(_single_compile, objects))
    return objects


# only if python version > 2.6, somehow the travis compiler hangs in 2.6
if sys.version_info[0] * 100 + sys.version_info[1] > 206:
    import distutils.ccompiler

    distutils.ccompiler.CCompiler.compile = parallelCCompile


# put a path in quotes if required
def quote_path(path):
    # looks like disutils don't need path quoting in version >= 3.9:
    if " " in path and sys.version_info[0] * 100 + sys.version_info[1] < 309:
        return '"' + path + '"'
    else:
        return path


# TODO: delete (Obsolete)
# patch get_ext_filename
from distutils.command.build_ext import build_ext

_old_get_ext_filename = build_ext.get_ext_filename


def patched_get_ext_filename(self, ext_name):
    r"""Convert the name of an extension (eg. "foo.bar") into the name
    of the file from which it will be loaded (eg. "foo/bar.so", or
    "foo\bar.pyd").
    """
    filename = _old_get_ext_filename(self, ext_name)
    # Making sure this matches qmake's default extension .dylib, instead of .so
    if platform.system() == "Darwin" and "_dbpi" in ext_name:
        filename = filename.replace(".so", ".dylib")
    return filename


distutils.command.build_ext.build_ext.get_ext_filename = patched_get_ext_filename

# end patch get_ext_filename


# patch CCompiler's library_filename for _dbpi libraries (SOs)
# Its default is to have .so for shared objects, instead of dylib,
# hence the patch

from distutils.ccompiler import CCompiler

old_library_filename = CCompiler.library_filename


def patched_library_filename(
    self, libname, lib_type="static", strip_dir=0, output_dir=""  # or 'shared'
):
    if platform.system() == "Darwin" and "_dbpi" in libname:
        lib_type = "dylib"
    return old_library_filename(
        self, libname, lib_type=lib_type, strip_dir=strip_dir, output_dir=output_dir
    )


distutils.ccompiler.CCompiler.library_filename = patched_library_filename

# end patch CCompiler's library_filename for _dbpi libraries (SOs)


# despite its name, setuptools.command.build_ext.link_shared_object won't
# link a shared object on Linux, but a static library and patches distutils
# for this ... We're patching this back now.


def always_link_shared_object(
    self,
    objects,
    output_libname,
    output_dir=None,
    libraries=None,
    library_dirs=None,
    runtime_library_dirs=None,
    export_symbols=None,
    debug=0,
    extra_preargs=None,
    extra_postargs=None,
    build_temp=None,
    target_lang=None,
):
    self.link(
        self.SHARED_LIBRARY,
        objects,
        output_libname,
        output_dir,
        libraries,
        library_dirs,
        runtime_library_dirs,
        export_symbols,
        debug,
        extra_preargs,
        extra_postargs,
        build_temp,
        target_lang,
    )


setuptools.command.build_ext.libtype = "shared"
setuptools.command.build_ext.link_shared_object = always_link_shared_object

# ----------------------------------------------------------------------------------------


class Config(object):

    """
    Provides some configuration-specific methods
    """

    def __init__(self):

        # TODO: is this really how we get the build paths?

        build_cmd = Distribution().get_command_obj("build")
        build_cmd.finalize_options()
        self.build_platlib = build_cmd.build_platlib
        self.build_temp = build_cmd.build_temp
        if platform.system() == "Windows":
            # Windows uses separate temp build folders for debug and release
            if build_cmd.debug:
                self.build_temp = os.path.join(self.build_temp, "Debug")
            else:
                self.build_temp = os.path.join(self.build_temp, "Release")

        build_ext_cmd = Distribution().get_command_obj("build_ext")

        build_ext_cmd.initialize_options()
        build_ext_cmd.setup_shlib_compiler()
        self.build_ext_cmd = build_ext_cmd

        self.root = "klayout"

    def add_extension(self, ext):
        self.build_ext_cmd.ext_map[ext.name] = ext

    def libname_of(self, mod, is_lib=None):
        """
        Returns the library name for a given module
        The library name is usually decorated (i.e. "tl" -> "tl.cpython-35m-x86_64-linux-gnu.so").
        This code was extracted from the source in setuptools.command.build_ext.build_ext
        """
        libtype = setuptools.command.build_ext.libtype
        full_mod = self.root + "." + mod
        if is_lib is True:
            # Here we are guessing it is a library and that the filename
            # is to be computed by shlib_compiler.
            # See source code of setuptools.command.build_ext.build_ext.get_ext_filename
            filename = self.build_ext_cmd.get_ext_filename(full_mod)
            fn, ext = os.path.splitext(filename)
            ext_path = self.build_ext_cmd.shlib_compiler.library_filename(fn, libtype)
        else:
            # This assumes that the Extension/Library object was added to the
            # ext_map dictionary via config.add_extension.
            assert full_mod in self.build_ext_cmd.ext_map
            ext_path = self.build_ext_cmd.get_ext_filename(full_mod)
        ext_filename = os.path.basename(ext_path)

        # Exception for database plugins, which will always be dylib.
        if platform.system() == "Darwin" and "_dbpi" in mod:
            ext_filename = ext_filename.replace(".so", ".dylib")
        return ext_filename

    def path_of(self, mod, mod_src_path):
        """
        Returns the build path of the library for a given module
        """
        if platform.system() == "Windows":
            # On Windows, the library to link is the import library
            (dll_name, dll_ext) = os.path.splitext(self.libname_of(mod))
            return os.path.join(self.build_temp, mod_src_path, dll_name + ".lib")
        else:
            return os.path.join(self.build_platlib, self.root, self.libname_of(mod))

    def compile_args(self, mod):
        """
        Gets additional compiler arguments
        """
        args: List[str]
        if platform.system() == "Windows":
            bits = os.getenv("KLAYOUT_BITS")
            if bits:
                args = [
                    quote_path("-I" + os.path.join(bits, "zlib", "include")),
                    quote_path("-I" + os.path.join(bits, "ptw", "include")),
                    quote_path("-I" + os.path.join(bits, "expat", "include")),
                    quote_path("-I" + os.path.join(bits, "curl", "include")),
                ]
            else:
                args = []
        else:
            args = [
                "-Wno-strict-aliasing",  # Avoids many "type-punned pointer" warnings
                "-std=c++11",  # because we use unordered_map/unordered_set
            ]
        return args

    def libraries(self, mod):
        """
        Gets the libraries to add
        """
        if platform.system() == "Windows":
            if mod == "_tl":
                return ["libcurl", "expat", "pthreadVCE2", "zlib", "wsock32"]
        else:
            if mod == "_tl":
                return ["curl", "expat"]
        return []

    def link_args(self, mod):
        """
        Gets additional linker arguments
        """
        if platform.system() == "Windows":
            args = ["/DLL"]
            bits = os.getenv("KLAYOUT_BITS")
            if bits:
                args += [
                    quote_path("/LIBPATH:" + os.path.join(bits, "zlib", "libraries")),
                    quote_path("/LIBPATH:" + os.path.join(bits, "ptw", "libraries")),
                    quote_path("/LIBPATH:" + os.path.join(bits, "expat", "libraries")),
                    quote_path("/LIBPATH:" + os.path.join(bits, "curl", "libraries")),
                ]
            return args
        elif platform.system() == "Darwin":
            # For the dependency modules, make sure we produce a dylib.
            # We can only link against such, but the bundles produced otherwise.
            args = []
            if mod[0] == "_":
                args += [
                    "-Wl,-dylib",
                    "-Wl,-install_name,@rpath/%s" % self.libname_of(mod, is_lib=True),
                ]
            args += ["-Wl,-rpath,@loader_path/"]
            return args
        else:
            # this makes the libraries suitable for linking with a path -
            # i.e. from path_of('_tl'). Without this option, the path
            # will be included in the reference and at runtime the loaded
            # will look for the path-qualified library. But that's the
            # build path and the loader will fail.
            args = []
            args += ["-Wl,-soname," + self.libname_of(mod, is_lib=True)]
            if "_dbpi" not in mod:
                loader_path = "$ORIGIN"
            else:
                loader_path = "$ORIGIN/.."
            args += ["-Wl,-rpath," + loader_path]
            # default linux shared object compilation uses the '-g' flag,
            # which generates unnecessary debug information
            # removing with strip-all during the linking stage
            args += ["-Wl,--strip-all"]
            return args

    def macros(self):
        """
        Returns the macros to use for building
        """
        return [
            ("HAVE_CURL", 1),
            ("HAVE_EXPAT", 1),
            ("KLAYOUT_MAJOR_VERSION", self.major_version()),
            ("KLAYOUT_MINOR_VERSION", self.minor_version()),
            ("GSI_ALIAS_INSPECT", 1),
        ]

    def minor_version(self):
        """
        Gets the version string
        """

        # this will obtain the version string from the "version.sh" file which
        # is the central point of configuration
        version_file = os.path.join(os.path.dirname(__file__), "version.sh")
        with open(version_file, "r") as file:
            version_txt = file.read()
            rm = re.search(
                r"KLAYOUT_VERSION\s*=\s*\"(.*?)\.(.*?)(\..*)?\".*", version_txt
            )
            if rm:
                version_string = rm.group(2)
                return version_string

        raise RuntimeError("Unable to obtain version string from version.sh")

    def major_version(self):
        """
        Gets the version string
        """

        # this will obtain the version string from the "version.sh" file which
        # is the central point of configuration
        version_file = os.path.join(os.path.dirname(__file__), "version.sh")
        with open(version_file, "r") as file:
            version_txt = file.read()
            rm = re.search(
                r"KLAYOUT_VERSION\s*=\s*\"(.*?)\.(.*?)(\..*)?\".*", version_txt
            )
            if rm:
                version_string = rm.group(1)
                return version_string

        raise RuntimeError("Unable to obtain version string from version.sh")

    def version(self):
        """
        Gets the version string
        """

        # this will obtain the version string from the "version.sh" file which
        # is the central point of configuration
        version_file = os.path.join(os.path.dirname(__file__), "version.sh")
        with open(version_file, "r") as file:
            version_txt = file.read()
            rm = re.search(r"KLAYOUT_PYPI_VERSION\s*=\s*\"(.*?)\".*", version_txt)
            if rm:
                version_string = rm.group(1)
                return version_string

        raise RuntimeError("Unable to obtain version string from version.sh")


config = Config()

# ------------------------------------------------------------------
# _tl dependency library

_tl_path = os.path.join("src", "tl", "tl")
_tl_sources = set(glob.glob(os.path.join(_tl_path, "*.cc")))

# Exclude sources which are compatible with Qt only
# Caveat, in source distribution tarballs from pypi, these files will
# not exist. So we need an error-free discard method instead of list's remove.
_tl_sources.discard(os.path.join(_tl_path, "tlHttpStreamQt.cc"))
_tl_sources.discard(os.path.join(_tl_path, "tlHttpStreamNoQt.cc"))
_tl_sources.discard(os.path.join(_tl_path, "tlFileSystemWatcher.cc"))
_tl_sources.discard(os.path.join(_tl_path, "tlDeferredExecutionQt.cc"))

_tl = Library(
    config.root + "._tl",
    define_macros=config.macros() + [("MAKE_TL_LIBRARY", 1)],
    language="c++",
    libraries=config.libraries("_tl"),
    extra_link_args=config.link_args("_tl"),
    extra_compile_args=config.compile_args("_tl"),
    sources=list(_tl_sources),
)

config.add_extension(_tl)

# ------------------------------------------------------------------
# _gsi dependency library

_gsi_path = os.path.join("src", "gsi", "gsi")
_gsi_sources = set(glob.glob(os.path.join(_gsi_path, "*.cc")))

_gsi = Library(
    config.root + "._gsi",
    define_macros=config.macros() + [("MAKE_GSI_LIBRARY", 1)],
    include_dirs=[_tl_path],
    extra_objects=[config.path_of("_tl", _tl_path)],
    language="c++",
    extra_link_args=config.link_args("_gsi"),
    extra_compile_args=config.compile_args("_gsi"),
    sources=list(_gsi_sources),
)
config.add_extension(_gsi)

# ------------------------------------------------------------------
# _pya dependency library

_pya_path = os.path.join("src", "pya", "pya")
_pya_sources = set(glob.glob(os.path.join(_pya_path, "*.cc")))

_pya = Library(
    config.root + "._pya",
    define_macros=config.macros() + [("MAKE_PYA_LIBRARY", 1)],
    include_dirs=[_tl_path, _gsi_path],
    extra_objects=[config.path_of("_tl", _tl_path), config.path_of("_gsi", _gsi_path)],
    language="c++",
    extra_link_args=config.link_args("_pya"),
    extra_compile_args=config.compile_args("_pya"),
    sources=list(_pya_sources),
)
config.add_extension(_pya)

# ------------------------------------------------------------------
# _db dependency library

_db_path = os.path.join("src", "db", "db")
_db_sources = set(glob.glob(os.path.join(_db_path, "*.cc")))

_db = Library(
    config.root + "._db",
    define_macros=config.macros() + [("MAKE_DB_LIBRARY", 1)],
    include_dirs=[_tl_path, _gsi_path, _db_path],
    extra_objects=[config.path_of("_tl", _tl_path), config.path_of("_gsi", _gsi_path)],
    language="c++",
    extra_link_args=config.link_args("_db"),
    extra_compile_args=config.compile_args("_db"),
    sources=list(_db_sources),
)
config.add_extension(_db)

# ------------------------------------------------------------------
# _lib dependency library

_lib_path = os.path.join("src", "lib", "lib")
_lib_sources = set(glob.glob(os.path.join(_lib_path, "*.cc")))

_lib = Library(
    config.root + "._lib",
    define_macros=config.macros() + [("MAKE_LIB_LIBRARY", 1)],
    include_dirs=[_tl_path, _gsi_path, _db_path, _lib_path],
    extra_objects=[
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_db", _db_path),
    ],
    language="c++",
    extra_link_args=config.link_args("_lib"),
    extra_compile_args=config.compile_args("_lib"),
    sources=list(_lib_sources),
)
config.add_extension(_lib)

# ------------------------------------------------------------------
# _rdb dependency library

_rdb_path = os.path.join("src", "rdb", "rdb")
_rdb_sources = set(glob.glob(os.path.join(_rdb_path, "*.cc")))

_rdb = Library(
    config.root + "._rdb",
    define_macros=config.macros() + [("MAKE_RDB_LIBRARY", 1)],
    include_dirs=[_db_path, _tl_path, _gsi_path],
    extra_objects=[
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_db", _db_path),
    ],
    language="c++",
    extra_link_args=config.link_args("_rdb"),
    extra_compile_args=config.compile_args("_rdb"),
    sources=list(_rdb_sources),
)
config.add_extension(_rdb)

# ------------------------------------------------------------------
# dependency libraries from db_plugins

db_plugins = []

dbpi_dirs = glob.glob(os.path.join("src", "plugins", "*", "db_plugin"))
dbpi_dirs += glob.glob(os.path.join("src", "plugins", "*", "*", "db_plugin"))

for pi in dbpi_dirs:

    mod_name = "_" + os.path.split(os.path.split(pi)[-2])[-1] + "_dbpi"

    pi_sources = glob.glob(os.path.join(pi, "*.cc"))

    pi_ext = Library(
        config.root + ".db_plugins." + mod_name,
        define_macros=config.macros() + [("MAKE_DB_PLUGIN_LIBRARY", 1)],
        include_dirs=[
            os.path.join("src", "plugins", "common"),
            _db_path,
            _tl_path,
            _gsi_path,
        ],
        extra_objects=[
            config.path_of("_tl", _tl_path),
            config.path_of("_gsi", _gsi_path),
            config.path_of("_db", _db_path),
        ],
        language="c++",
        extra_link_args=config.link_args(mod_name),
        extra_compile_args=config.compile_args(mod_name),
        sources=pi_sources,
    )

    db_plugins.append(pi_ext)
    config.add_extension(pi_ext)

# ------------------------------------------------------------------
# tl extension library

tl_path = os.path.join("src", "pymod", "tl")
tl_sources = set(glob.glob(os.path.join(tl_path, "*.cc")))

tl = Extension(
    config.root + ".tlcore",
    define_macros=config.macros(),
    include_dirs=[_tl_path, _gsi_path, _pya_path],
    extra_objects=[
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_pya", _pya_path),
    ],
    extra_link_args=config.link_args("tlcore"),
    extra_compile_args=config.compile_args("tlcore"),
    sources=list(tl_sources),
)

# ------------------------------------------------------------------
# db extension library

db_path = os.path.join("src", "pymod", "db")
db_sources = set(glob.glob(os.path.join(db_path, "*.cc")))

db = Extension(
    config.root + ".dbcore",
    define_macros=config.macros(),
    include_dirs=[_db_path, _tl_path, _gsi_path, _pya_path],
    extra_objects=[
        config.path_of("_db", _db_path),
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_pya", _pya_path),
    ],
    extra_link_args=config.link_args("dbcore"),
    extra_compile_args=config.compile_args("dbcore"),
    sources=list(db_sources),
)

# ------------------------------------------------------------------
# lib extension library

lib_path = os.path.join("src", "pymod", "lib")
lib_sources = set(glob.glob(os.path.join(lib_path, "*.cc")))

lib = Extension(
    config.root + ".libcore",
    define_macros=config.macros(),
    include_dirs=[_lib_path, _tl_path, _gsi_path, _pya_path],
    extra_objects=[
        config.path_of("_lib", _lib_path),
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_pya", _pya_path),
    ],
    extra_link_args=config.link_args("libcore"),
    extra_compile_args=config.compile_args("libcore"),
    sources=list(lib_sources),
)

# ------------------------------------------------------------------
# rdb extension library

rdb_path = os.path.join("src", "pymod", "rdb")
rdb_sources = set(glob.glob(os.path.join(rdb_path, "*.cc")))

rdb = Extension(
    config.root + ".rdbcore",
    define_macros=config.macros(),
    include_dirs=[_rdb_path, _tl_path, _gsi_path, _pya_path],
    extra_objects=[
        config.path_of("_rdb", _rdb_path),
        config.path_of("_tl", _tl_path),
        config.path_of("_gsi", _gsi_path),
        config.path_of("_pya", _pya_path),
    ],
    extra_link_args=config.link_args("rdbcore"),
    extra_compile_args=config.compile_args("rdbcore"),
    sources=list(rdb_sources),
)

# ------------------------------------------------------------------
# Core setup function

if __name__ == "__main__":
    setup(
        name=config.root,
        version=config.version(),
        license="GNU GPLv3",
        description="KLayout standalone Python package",
        long_description="This package is a standalone distribution of KLayout's Python API.\n\nFor more details see here: https://www.klayout.org/klayout-pypi",
        author="Matthias Koefferlein",
        author_email="matthias@klayout.de",
        classifiers=[
            # Recommended classifiers
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            # Optional classifiers
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        ],
        url="https://github.com/klayout/klayout",
        packages=find_packages("src/pymod/distutils_src"),
        package_dir={
            "": "src/pymod/distutils_src"
        },  # https://github.com/pypa/setuptools/issues/230
        package_data={config.root: ["src/pymod/distutils_src/klayout/*.pyi"]},
        include_package_data=True,
        ext_modules=[_tl, _gsi, _pya, _db, _lib, _rdb]
        + db_plugins
        + [tl, db, lib, rdb],
    )
